<?php

namespace WPIS;

require_once 'class-api-public.php';
require_once 'class-wpis-post-types.php';
require_once 'class-wpis-cache.php';

require_once  ABSPATH . 'wp-admin/includes/image.php';

class ImportEstates {

    private $cacheFile;
    private $postTypeEstateName;
    private $postTypeProjectName;
    private $postTypeAgentName;
    private $debug = false;

    private $enableImportSystem = false;
    private $enableImportCustomPostType = false;
    private $enableProjectManagement = false;
    private $enableProjectManagementCustomPostType = false;
    private $enableAgentManagement = false;
    private $enableAgentManagementCustomPostType = false;

    private $enableTaxonomies = false;
    private $autoRemoveUpdatedEstates = false;
    private $expiredLockTime = 1800;
    private $cacheDirectory;
    public  $cache;

    static public function init() {

        error_reporting(E_ALL);
        ini_set('display_errors', 1);

        $class = new ImportEstates();

        $class->cacheDirectory = wp_get_upload_dir()['basedir'] . '/wp-immo-sync';

        if( ! is_dir($class->cacheDirectory) ){
            mkdir($class->cacheDirectory);
        }

        $class->cache = new Cache($class->cacheDirectory . '/' . 'cache');

        $class->wp_db = $GLOBALS['wpdb'];

        $class->cacheFile = $class->cacheDirectory . '/wpis-cache.php';
        $class->lockFile = $class->cacheDirectory . '/wpis.lock';
        $class->logFile = $class->cacheDirectory . '/wpis-import.log';

        $class->checkSettings();

        if( $class->enableImportSystem && !$class->enableImportCustomPostType ){
            $class->postTypeEstateName = PostTypes::$postTypeEstateName;
        }
        elseif( $class->enableImportSystem && $class->enableImportCustomPostType ){
            $class->postTypeEstateName = SettingsManagement::get('customPostType');
        }

        if( $class->enableProjectManagement && !$class->enableProjectManagementCustomPostType ){
            $class->postTypeProjectName = PostTypes::$postTypeProjectName;
        }
        elseif( $class->enableProjectManagement && $class->enableProjectManagementCustomPostType ){
            $class->postTypeProjectName = SettingsManagement::get('projectCustomPostType');
        }

        if( $class->enableAgentManagement && !$class->enableAgentManagementCustomPostType ){
            $class->postTypeAgentName = PostTypes::$postTypeAgentName;
        }
        elseif( $class->enableAgentManagement && $class->enableAgentManagementCustomPostType ){
            $class->postTypeAgentName = SettingsManagement::get('agentCustomPostType');
        }

        return $class;
    }

    private function checkSettings(){
        $this->enableImportSystem = SettingsManagement::get('enableImportSystem');
        $this->enableImportCustomPostType = SettingsManagement::get('enableImportCustomPostType');
        $this->autoRemoveUpdatedEstates = SettingsManagement::get('autoRemoveUpdatedEstates');
        $this->enableTaxonomies = SettingsManagement::get('enableTaxonomies');
        $this->enableProjectManagement               = settingsManagement::get('enableProjectManagement');
        $this->enableProjectManagementCustomPostType = settingsManagement::get('enableProjectManagementCustomPostType');
        $this->enableAgentManagement               = settingsManagement::get('enableAgentManagement');
        $this->enableAgentManagementCustomPostType = settingsManagement::get('enableAgentManagementCustomPostType');
    }

    public function checking() : array {

        $this->log('[IMPORT] CHECKING START');
        $limit = 350;
        $start = microtime(true);

        $requestP1 = ApiPublic::init()->request('GET', 'estates', ['limit' => $limit]);

        if( json_decode($requestP1->data)->total === 0 ){
            return [
                'success' => false,
                'message' =>'There is no estates to import.',
                'code' => 1000
            ];
        }

        if(  $requestP1->code !== 200 || !json_decode($requestP1->data)->data){
            $this->log('[IMPORT] CHECKING ERROR');
            $this->log('[IMPORT] CODE : ' . $requestP1->code);
            $this->log('[IMPORT] RESPONSE : ' . $requestP1->data);
            return [
                'success' => false,
                'message' =>'An error occured while retrieving estates.',
                'code' => 1001,
                'data' => $requestP1->data
            ];
        }

        // Get next page if number of estates exceeded maximum limit
        $response = json_decode($requestP1->data);
        $estates = $response->data;
        if( $response->total > $limit ) {
            $requestP2 = ApiPublic::init()->request('GET', 'estates', ['limit' => $limit, 'start' => $limit]);
            if( $requestP2->code !== 200 || !json_decode($requestP2->data)->data){
                $this->log('[IMPORT] CHECKING ERROR');
                $this->log('[IMPORT] CODE : ' . $requestP2->code);
                $this->log('[IMPORT] RESPONSE : ' . $requestP2->data);

                return [
                    'success' => false,
                    'message' =>'An error occured while retrieving estates.',
                    'data' => $requestP2->data
                ];
            }
            $response = json_decode($requestP2->data);
            $estates = array_merge($estates, $response->data);
        }

        $query =
        'SELECT ' .
        '   id as postId, ' .
        '   meta_value as estateId, '.
        '   (SELECT meta_value FROM ' . $this->wp_db->prefix . 'postmeta WHERE ' . $this->wp_db->prefix . 'postmeta.post_id = ' . $this->wp_db->prefix . 'posts.id AND meta_key = "wpis_locked" LIMIT 1) as locked, '.
        '   (SELECT CASE WHEN meta_value IS NULL OR meta_value = "" THEN false ELSE true END FROM ' . $this->wp_db->prefix . 'postmeta WHERE ' . $this->wp_db->prefix . 'postmeta.post_id = ' . $this->wp_db->prefix . 'posts.id AND meta_key = "wpis_signatureChanged" LIMIT 1) as signatureChanged '.
        'FROM ' . $this->wp_db->prefix . 'posts ' .
        'JOIN ' . $this->wp_db->prefix . 'postmeta ON ' . $this->wp_db->prefix . 'postmeta.post_id = ' . $this->wp_db->prefix . 'posts.id ' .
        'WHERE post_type IN ("' . $this->postTypeEstateName . '", "' . $this->postTypeProjectName . '") ' .
        'AND meta_key = "' . PostTypes::$postTypeMetaPrefix . '_id"';

        $estatesImported = $this->wp_db->get_results($query);
        $estatesImportedMap = [];

        // Map estate imported in associative array
        foreach($estatesImported as $estateImported){
            $estatesImportedMap[$estateImported->estateId] = $estateImported;
        }

        $totalImported = count($estatesImportedMap);
        $totalToImport = $response->total;
        $estatesMapped = [];

        // Map data from API with imported estates
        foreach( $estates as $estate ){
            $estatesMapped[] = [
                'id'             => $estate->id,
                'imported'       => isset($estatesImportedMap[$estate->id]),
                'locked'         => isset($estatesImportedMap[$estate->id]) && $estatesImportedMap[$estate->id]->locked,
                'hasBeenUpdated' => isset($estatesImportedMap[$estate->id]) && $estatesImportedMap[$estate->id]->signatureChanged,
                'data'           => $estate
            ];
        }

        // Create cache
        $cache = ['estates'  => $estatesMapped];
        $this->cache->set('data', $cache);

        // Clear estates imported of from estates from api to determine wich one should be removed
        foreach($estatesImportedMap as $estateId => $estate ){
            foreach($estatesMapped as $estateFromApi ){

                if( $estate->locked )
                    unset($estatesImportedMap[$estateId]);

                if( $this->autoRemoveUpdatedEstates && $estateFromApi['hasBeenUpdated'] )
                    continue;

                if( (string) $estateId === (string) $estateFromApi['id'] )
                    unset($estatesImportedMap[$estateId]);

            }
        }

        $estateToRemove = $estatesImportedMap;

        // Generate inactive list
        if( !empty($estateToRemove) ){
            foreach( $estateToRemove as $key => $estate ){
                $type = get_post_meta( $estate->postId , PostTypes::$postTypeMetaPrefix . '_' . 'category_label' , true );
                $category = get_post_meta( $estate->postId , PostTypes::$postTypeMetaPrefix . '_' . 'subcategory_label' , true );
                $purpose = get_post_meta( $estate->postId, PostTypes::$postTypeMetaPrefix . '_' . 'purpose_label', true );
                $typeCategory = PostTypes::ucfirstUtf8($type) . ' / ' . PostTypes::ucfirstUtf8($category);
                $price = round(get_post_meta( $estate->postId , PostTypes::$postTypeMetaPrefix . '_' . 'finance_price' , true ));
                $price = number_format_i18n($price) . ' €';
                $price = str_replace(',', ' ', $price);

                $estateToRemove[$key]->postTitle = get_post_meta( $estate->postId, PostTypes::$postTypeMetaPrefix . '_' . 'name', true );
                $estateToRemove[$key]->reference = get_post_meta( $estate->postId, PostTypes::$postTypeMetaPrefix . '_' . 'reference', true );
                $estateToRemove[$key]->price = $price;
                $estateToRemove[$key]->category = $typeCategory;
                $estateToRemove[$key]->purpose = PostTypes::ucfirstUtf8($purpose);
                $estateToRemove[$key]->postDateImported = get_the_time('d-m-Y', $estate->postId);
            }
        }

        $this->log('[IMPORT] CHECKING RESULT : imported: ' . $totalImported . ', total: ' . $totalToImport . ', inactives: ' . count($estateToRemove) . ', time: ' . round(microtime(true) - $start, 3) );

        return [
            'success' => true,
            'totalImported' => $totalImported,
            'estateTotal' => $totalToImport,
            'estatesInactive' => $estateToRemove
        ];
    }

    public function next( $settings ){

        if( $this->isLocked() ){
            return [
                'success' => false,
                'message' => 'An import is currently running',
            ];
        }

        $this->lock();

        if( $this->cache->get('data') === false ){
            $this->checking();
        }

        // Initialise status
        if( ! $cachedData = $this->cache->get('data') ){

            $this->unlock();
            $this->log('[IMPORT][ERROR] The cache is empty.');

            return [
                'success'           => false,
                'locked'            => false,
            ];
        }

        if( ! isset($cachedData['estates']) && ! is_array($cachedData['estates']) ){

            $this->unlock();
            $this->log('[IMPORT][ERROR] Estates cached are corrupted : ' . json_encode($cachedData));

            return [
                'success'           => false,
                'locked'            => false,
            ];
        }

        $estateImported = 0;
        $totalEstates = count($cachedData['estates']);
        $startTime = microtime(true);
        $isComplete = false;

        // Count how many estates are imported
        foreach( $cachedData['estates'] as $index => $estate ){
            if( $estate['imported'] ){
                $estateImported++;
            }
        }

        // Determine estate to import or update
        $targetEstate = null;
        foreach( $cachedData['estates'] as $index => $estate ){
            if( !$estate['imported'] ){
                $targetIndex = $index;
                $targetEstate = $estate;
                break;
            }
        }

        if( ! is_null($targetEstate) ){
            // Import or update targeted estate
            $forceImportImages = isset($settings->forceImportImages) ? $settings->forceImportImages : false;
            $origin = isset($settings->origin) ? $settings->origin : 'admin';

            if( ! $lastImport = $this->importEstate( $targetEstate['data'], false, $forceImportImages, $origin ) )
            {
                $this->unlock();
                return [
                    'success' => false ,
                    'message' => 'An error occured while importing estates.'
                ];
            }

            // Update cache data
            $targetEstate['imported'] = true;
            $targetEstate['locked'] = false;

            $cachedData['estates'][$targetIndex] = $targetEstate;

            $this->cache->set('data', $cachedData);
        }
        else {
            $isComplete = true;
            $lastImport = null;
        }

        $this->unlock();

        $response = [
            'success'           => true,
            'locked'            => false,
            'isComplete'        => $isComplete,
            'time'              => round(microtime(true) - $startTime, 3),
            'lastImport'        => $lastImport,
            'estateTotal'       => $totalEstates,
            'estateImported'    => $estateImported,
        ];


        return $response;
    }

    public function updateStart() : array {

        $start = microtime(true);

        if( $this->isLocked() ){
            return [
                'success' => false,
                'message' => 'An import is currently running',
            ];
        }

        $this->lock();
        $this->checking();

        $data = $this->cache->get('data');

        if( is_null( $data['estates'] ) ){
            $elapsedTime = round(microtime(true) - $start, 3);
            $this->log('[IMPORT] ERROR NO ESTATES AVAILABLES');
            return [
                'success' => false,
                'message' => 'An error occured while reading cache.',
                'time' => $elapsedTime,
            ];
        }

        $progress = [
            'current' => 0,
            'total' => count($data['estates']),
            'isProcessing' => true
        ];

        foreach( $data['estates'] as $index => $estate ){
            if( $this->importEstate( $estate['data'], true ) ){
                $progress['current'] = $progress['current'] + 1;
                $this->cache->set('progress', $progress );
            }
        }

        $progress['isProcessing'] = false;

        $this->cache->set('progress', $progress );
        $this->cache->delete('data');

        $elapsedTime = round(microtime(true) - $start, 3);

        $this->unlock();
        $this->log('[IMPORT] UPDATE estatesUpdatedCount: ' . $progress['current'] . ', time: ' . $elapsedTime );

        return [
            'success' => true,
            'time' => $elapsedTime,
            'estatesUpdatedCount' => $progress['current']
        ];

    }

    public function updateProgress() {

        $progress = $this->cache->get('progress');

        if( $progress && !$progress['isProcessing'] || $progress['current'] === $progress['total'] ){
            $this->cache->delete('progress');
        }

        return $progress;
    }

    public function importEstate( $estate, $updateOnly = false, $forceImportImages = false ){

        $start = microtime(true);
        $isUpdate = false;
        $postFound = $this->wp_db->get_results("SELECT post_id FROM " . $this->wp_db->prefix . "postmeta WHERE meta_key = 'wpis_id' AND meta_value = '" . $estate->id . "' LIMIT 1");

        if( !empty( $postFound ) ){
            // UPDATE
            $isUpdate = true;
            $postId = $postFound[0]->post_id;
            $post = [
                'ID'           => $postId,
                'post_title'   => $estate->name,
                'post_content' => $estate->description->base,
            ];

            wp_update_post( $post );

            if( $forceImportImages ){
                $this->importPictures($estate, $postId);
            }
        }
        elseif( !$updateOnly ) {
            // Create
            $postId = wp_insert_post([
                'post_type'      => $this->postTypeEstateName,
                'post_title'     => $estate->name,
                'post_content'   => $estate->description->base,
                'post_status'    => 'publish',
                'comment_status' => 'closed',
                'ping_status'    => 'closed'
            ]);

            update_post_meta( $postId, PostTypes::$postTypeMetaPrefix . '_locked',  0);

            $this->importPictures($estate, $postId);
        }
        else {
            return false;
        }

        foreach( PostTypes::$postTypeMetas as $group){
            foreach($group['fields'] as $property){

                $propertySegments = explode('.', $property['field']);
                $postMetaName = PostTypes::$postTypeMetaPrefix . '_' . str_replace('.', '_', $property['field']);

                if( property_exists($estate, $propertySegments[0]) ){
                    if( count($propertySegments) > 1 ){
                        if( is_object($estate->{$propertySegments[0]}) && property_exists($estate->{$propertySegments[0]}, $propertySegments[1]) ){
                            update_post_meta( $postId, $postMetaName,  $estate->{$propertySegments[0]}->{$propertySegments[1]});
                        }
                    }
                    else {
                        update_post_meta( $postId, $postMetaName,  $estate->{$propertySegments[0]});
                    }
                }
            }
        }

        // Generate signature
        $newSignature = md5(json_encode($estate->pictures));
        $oldSignature = get_post_meta($postId, PostTypes::$postTypeMetaPrefix . '_signature', true);

        if( get_post_meta($postId, PostTypes::$postTypeMetaPrefix . '_signatureChanged', true) !== '1' ){
            update_post_meta( $postId, PostTypes::$postTypeMetaPrefix . '_signatureChanged' , $oldSignature && $newSignature !== $oldSignature );
        }

        update_post_meta( $postId, PostTypes::$postTypeMetaPrefix . '_signature' ,  $newSignature );

        // Taxonomies mapping
        if( $this->enableTaxonomies ){
            // > estate_purposes
            $estatePurposeLabel = get_post_meta($postId, 'wpis_purpose_label', true);
            if( ! $term = term_exists( $estatePurposeLabel, 'estate_purposes' ) ){
                $term = wp_insert_term( $estatePurposeLabel, 'estate_purposes');
            }
            if( !is_wp_error($term) ){
                wp_set_post_terms( $postId, $term['term_id'], 'estate_purposes');
            }

            // > estate_cities
            $estateCity = get_post_meta($postId, 'wpis_address_city', true);
            if( ! $term = term_exists( $estateCity, 'estate_cities' ) ){
                $term = wp_insert_term( $estateCity, 'estate_cities');
            }
            if( !is_wp_error($term) ){
                wp_set_post_terms( $postId, $term['term_id'], 'estate_cities');
            }

            // > estate_categories
            $estateCategory = get_post_meta($postId, 'wpis_category_label', true);
            if( ! $term = term_exists( $estateCategory, 'estate_categories' ) ){
                $term = wp_insert_term( ucfirst($estateCategory), 'estate_categories');
            }
            if( !is_wp_error($term) ){
                wp_set_post_terms( $postId, $term['term_id'], 'estate_categories');
            }

            // > estate_ameneties
            $estateAmenetiesGeneric = get_post_meta($postId, 'wpis_ameneties_generic', true);
            $estateAmenetiesCustom = get_post_meta($postId, 'wpis_ameneties_custom', true);
            $estateAmeneties = array_merge($estateAmenetiesGeneric, $estateAmenetiesCustom);
            $amenetiesTermIds = [];

            foreach( $estateAmeneties as $ammenety ){
                if( ! $term = term_exists( $ammenety, 'estate_ammeneties' ) ){
                    $term = wp_insert_term( $ammenety, 'estate_ammeneties' );
                }
                if( !is_wp_error($term) ){
                    $amenetiesTermIds[] = (int) $term['term_id'];
                }
            }

            wp_set_post_terms( $postId, $amenetiesTermIds, 'estate_ammeneties');
        }

        // Split estate and project if project management enabled
        if( $this->enableProjectManagement ){
            if( get_post_meta($postId, 'wpis_properties_hasChildren', true ) ) {

                wp_update_post(['ID' => $postId, 'post_type' => $this->postTypeProjectName]);

                $estateId = get_post_meta($postId, 'wpis_id', true );
                $childPosts = query_posts([
                    'post_type' => $this->postTypeEstateName,
                    'posts_per_page' => -1,
                    'meta_query' => [
                        [
                          'key'     => PostTypes::$postTypeMetaPrefix . '_' . 'properties_parentId',
                          'compare' => '=',
                          'value'   => $estateId,
                          'type'    => 'CHAR',
                        ]
                    ]
                ]);

                foreach( $childPosts as $childPost ){
                    wp_update_post(['ID' => $childPost->ID, 'post_parent' => $postId]);
                }

            }
        }

        // Import agent if agent management enabled
        if( $this->enableAgentManagement ){
            if( get_post_meta($postId, 'wpis_agent_firstname', true ) ) {

                // wp_update_post(['ID' => $postId, 'post_type' => $this->postTypeProjectName]);

                $agentFirstname = get_post_meta($postId, 'wpis_agent_firstname', true );
                $agentLastname  = get_post_meta($postId, 'wpis_agent_lastname', true );
                $agentPicture   = get_post_meta($postId, 'wpis_agent_picture', true );
                $agentEmail     = get_post_meta($postId, 'wpis_agent_email', true );
                $agentPhone     = get_post_meta($postId, 'wpis_agent_phone', true );
                $agentRole      = get_post_meta($postId, 'wpis_agent_role', true );

                $agentPost = get_posts([
                    'post_type'  => $this->postTypeAgentName,
                    'title' => $agentFirstname . ' ' . $agentLastname,
                ]);

                $agentPostId = isset($agentPost[0]) ? $agentPost[0]->ID : 0;

                $postId = wp_insert_post([
                    'ID' => $agentPostId,
                    'post_title' => $agentFirstname . ' ' . $agentLastname,
                    'post_status' => 'publish',
                    'post_type' => $this->postTypeAgentName
                ]);

                update_post_meta($postId, 'wpis_agent_picture', $agentPicture, true );
                update_post_meta($postId, 'wpis_agent_firstname', $agentFirstname, true );
                update_post_meta($postId, 'wpis_agent_lastname', $agentLastname, true );
                update_post_meta($postId, 'wpis_agent_email', $agentEmail, true );
                update_post_meta($postId, 'wpis_agent_phone', $agentPhone, true );
                update_post_meta($postId, 'wpis_agent_role', $agentRole, true );

            }
        }

        // Post Update
        if( file_exists(get_stylesheet_directory() . '/wpis-post-update.php' ) ){
            require_once get_stylesheet_directory() . '/wpis-post-update.php';
            if( function_exists('wpis_post_update') ) {
                wpis_post_update($postId);
            }
        }

        // Logging
        if( $postId ){
            $this->log('[IMPORT] ' . ($isUpdate ? 'UPDATE' : 'CREATE' ) . ' postId: ' . $postId . ', estateId: ' . $estate->id . ', imageCount: ' . count($estate->pictures) . ', time: ' . round((microtime(true) - $start), 3 ));
        }

        return [
            'postId' => $postId,
            'estateId' => $estate->id,
            'estateName' => $estate->name
        ];
    }

    public function importPictures($estate, $postId){

        $wpUploadDir = wp_upload_dir();
        $hasPostThumbnail = false;

        if( $wpUploadDir['error'] ){
            return $wpUploadDir['error'];
        }

        foreach($estate->pictures as $index => $picture ){

            $filename   = $estate->id . '_' . $picture->id . '.' . $picture->format;
            $realpath   = $wpUploadDir['path'] . '/' . $filename;
            $url        = $wpUploadDir['url'] . '/' . $filename;

            $url = $picture->urls->large;
            $ch = curl_init();
            curl_setopt($ch, CURLOPT_URL, $url);
            curl_setopt($ch, CURLOPT_HEADER, false);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
            curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);

            $data = curl_exec($ch);
            $code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
            curl_close($ch) ;

            if( $code === 200 ){
                $result = file_put_contents($realpath, $data);

                $uploadId = wp_insert_attachment([
                    'guid'           => $realpath,
                    'post_mime_type' => mime_content_type( $realpath ),
                    'post_title'     => $filename,
                    'post_content'   => '',
                    'post_status'    => 'inherit',
                    'menu_order'     => $index,
                ], $realpath,  $postId, true);

                if( is_wp_error($uploadId) ){
                    $this->log('[IMPORT] IMAGE INSERT ATACHMENT ERROR :');
                    $this->log('[IMPORT] URL : ' . $url);
                    $this->log('[IMPORT] REALPATH : ' . $realpath);
                    $this->log('[IMPORT] POST_ID : ' . $postId);
                    $this->log('[IMPORT] WP_ERROR : ' . json_encode($uploadId));
                }

                wp_update_attachment_metadata( $uploadId, wp_generate_attachment_metadata( $uploadId, $realpath ) );

                if( !$hasPostThumbnail ){
                    set_post_thumbnail($postId, $uploadId);
                    $hasPostThumbnail = true;
                }
            }
            else {
                $this->log('[IMPORT] IMAGE IMPORT ERROR : ' . $code . ' ' . $data);
                $this->log('[IMPORT] URL : ' . $url);
                $this->log('[IMPORT] CODE : ' . $code);
                $this->log('[IMPORT] RESPONSE : ' . $data);
            }

        }

        return 'success';
    }

    public function removeInactives() : array {
        $checking = $this->checking();

        if( isset($checking['success']) && !$checking['success']){
            return $checking;
        }

        $i = 0;

        foreach($checking['estatesInactive'] as $estate ){
            $attachments = get_attached_media( '', $estate->postId );
            foreach ($attachments as $attachment) {
                wp_delete_attachment($attachment->ID, true);
            }
            wp_delete_post($estate->postId, true);
            $i++;
        }

        return [
            'success' => true,
            'estatesRemoved' => $i,
            'estatesRemovedList' => $checking['estatesInactive']
        ];
    }

    public function removeAll() : array {

        $query = new \WP_Query( [
            'post_type' => [$this->postTypeEstateName, $this->postTypeProjectName],
            'meta_key' => 'wpis_id',
            'posts_per_page' => 1000
        ]);

        $i = 0;

        if( !$query->have_posts() ){
            return [
                'success' => false,
                'message' => 'There is no estate to remove.'
            ];
        }

        if( $this->enableTaxonomies ){
            foreach( ['estate_cities', 'estate_categories', 'estate_purposes', 'estate_ammeneties'] as $taxonomy ){
                $terms = get_terms( $taxonomy, ['fields' => 'ids', 'hide_empty' => false] );
                foreach ( $terms as $value ) {
                    wp_delete_term( $value, $taxonomy );
                }
            }
        }

        foreach( $query->posts as $post ){
            $attachments = get_attached_media( '', $post->ID );
            foreach ($attachments as $attachment) {
                wp_delete_attachment($attachment->ID, true);
            }
            wp_delete_post($post->ID, true);
            $i++;
        }

        return [
            'success' => true,
            'estatesRemoved' => $i,
        ];
    }

    private function isLocked() : bool {

        if( $this->debug ) {
            return false;
        }

        if( file_exists($this->lockFile) ){

            if( (new \DateTime())->getTimestamp() - filemtime($this->lockFile) >= $this->expiredLockTime ){
                $this->log('[IMPORT] LOCK FILE HAS BEEN REMOVED DUE TO IT LIFE TIME.');
                $this->unlock();
                return $this->isLocked();
            }

            return true;
        }

        return false;
    }

    private function lock() : bool {

        if( $this->debug ) {
            return false;
        }

        touch($this->lockFile);

        return true;
    }

    private function unlock() : bool {

        if( $this->debug ) {
            return true;
        }

        if( file_exists($this->lockFile) ){
            unlink($this->lockFile);
        }

        return true;
    }

    private function log( $message ) : void {
        $date = (new \DateTime('NOW'))->format("y:m:d H:i:s");
        $logLine = "\r\n" . "{$date} " . $message;
        file_put_contents($this->logFile, $logLine, FILE_APPEND);
    }

}